@@ -3,7 +3,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2018	PgPool Global Development Group
+ * Copyright (c) 2003-2019	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -187,9 +187,18 @@ connection_do_auth(POOL_CONNECTION_POOL_SLOT * cp, char *password)
 
 		buf = palloc0(2 * (MD5_PASSWD_LEN + 4));	/* hash + "md5" + '\0' */
 
-		/* build md5 password */
+		/* set buffer address for building md5 password */
 		buf1 = buf + MD5_PASSWD_LEN + 4;
-		pool_md5_encrypt(password, cp->sp->user, strlen(cp->sp->user), buf1);
+
+		/*
+		 * If the supplied password is already in md5 hash format, we just
+		 * copy it. Otherwise calculate the md5 hash value.
+		 */
+		if (!strncmp("md5", password, 3) && (strlen(password) - 3) == MD5_PASSWD_LEN)
+			memcpy(buf1, password + 3, MD5_PASSWD_LEN + 1);
+		else
+			pool_md5_encrypt(password, cp->sp->user, strlen(cp->sp->user), buf1);
+
 		pool_md5_encrypt(buf1, salt, 4, buf + 3);
 		memcpy(buf, "md5", 3);
 
@@ -484,7 +493,7 @@ pool_do_auth(POOL_CONNECTION * frontend, POOL_CONNECTION_POOL * cp)
 		/*
 		 * check if we can use md5 authentication.
 		 */
-		if (NUM_BACKENDS > 1)
+		if (!RAW_MODE && NUM_BACKENDS > 1)
 		{
 			if (get_auth_password(MASTER(cp), frontend, 0,
 								  &password, &passwordType) == false)
@@ -1010,12 +1019,35 @@ do_clear_text_password(POOL_CONNECTION * backend, POOL_CONNECTION * frontend, in
 		/* frontend and backend are both authenticated already */
 		return 0;
 	}
+
 	if (get_auth_password(backend, frontend, reauth, &pwd, &passwordType) == false)
 	{
-		ereport(FATAL,
-				(return_code(2),
-				 errmsg("clear text password authentication failed"),
-				 errdetail("unable to get the password for user: \"%s\"", frontend->username)));
+		/*
+		 * We do not have any passeord, we can still get the password
+		 * from client using plain text authentication if it is
+		 * allowed by user
+		 */
+
+		if (frontend->pool_hba == NULL ||
+			frontend->pool_hba->auth_method == uaPassword ||
+			pool_config->allow_clear_text_frontend_auth )
+		{
+			ereport(DEBUG1,
+				(errmsg("using clear text authentication with frontend"),
+					 errdetail("backend is using password authentication")));
+
+			authenticate_frontend_clear_text(frontend);
+
+			/* now check again if we have a password now */
+
+			if (get_auth_password(backend, frontend, reauth, &pwd, &passwordType) == false)
+			{
+				ereport(FATAL,
+						(return_code(2),
+							errmsg("clear text password authentication failed"),
+							errdetail("unable to get the password for user: \"%s\"", frontend->username)));
+			}
+		}
 	}
 
 	if (passwordType == PASSWORD_TYPE_AES)
@@ -1042,7 +1074,7 @@ do_clear_text_password(POOL_CONNECTION * backend, POOL_CONNECTION * frontend, in
 		return 0;
 	}
 
-	size = pwd ? strlen(pwd) : 0;
+	size = strlen(pwd);
 
 	/* connection reusing? */
 	if (reauth)
@@ -1441,8 +1473,6 @@ authenticate_frontend(POOL_CONNECTION * frontend)
 			break;
 #ifdef USE_PAM
 		case uaPAM:
-			ereport(ERROR,
-					(errmsg("authenticate_frontend called with PAM")));
 			break;
 #endif							/* USE_PAM */
 
@@ -1707,7 +1737,7 @@ do_md5(POOL_CONNECTION * backend, POOL_CONNECTION * frontend, int reauth, int pr
 	char		encbuf[POOL_PASSWD_LEN + 1];
 	char	   *pool_passwd = NULL;
 
-	if (NUM_BACKENDS == 1)
+	if (RAW_MODE || NUM_BACKENDS == 1)
 		return do_md5_single_backend(backend, frontend, reauth, protoMajor);
 
 	if (passwordType == PASSWORD_TYPE_AES)
