@@ -1,5 +1,5 @@
 /*
-FILE * $Header$
+ * $Header$
  *
  * Handles PCP connection, and protocol communication with pgpool-II
  * These are client APIs. Server program should use APIs in pcp_stream.c
@@ -8,7 +8,7 @@ FILE * $Header$
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2016	PgPool Global Development Group
+ * Copyright (c) 2003-2019	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -729,6 +729,20 @@ process_node_info_response(PCPConnInfo * pcpConn, char *buf, int len)
 		if (index == NULL)
 			goto INVALID_RESPONSE;
 
+		index++;
+		strlcpy(backend_info->replication_state, index, sizeof(backend_info->replication_state));
+
+		index = (char *) memchr(index, '\0', len);
+		if (index == NULL)
+			goto INVALID_RESPONSE;
+
+		index++;
+		strlcpy(backend_info->replication_sync_state, index, sizeof(backend_info->replication_sync_state));
+
+		index = (char *) memchr(index, '\0', len);
+		if (index == NULL)
+			goto INVALID_RESPONSE;
+		
 		index++;
 		backend_info->status_changed_time = atol(index);
 
@@ -824,6 +838,7 @@ process_process_count_response(PCPConnInfo * pcpConn, char *buf, int len)
 				pcp_internal_error(pcpConn,
 								   "command failed. invalid response");
 				setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				pfree(process_list);
 				return;
 			}
 			index += 1;
