@@ -278,12 +278,11 @@ main(int argc, char **argv)
 		{
 			if (kill(pid, 0) == 0)
 			{
-				ereport(FATAL,
-						(errmsg("pid file found. is another pgpool(%d) is running?\n", pid)));
+				fprintf(stderr, "ERROR: pid file found. is another pgpool(%d) is running?\n", pid);
+				exit(EXIT_FAILURE);
 			}
 			else
-				ereport(NOTICE,
-						(errmsg("pid file found but it seems bogus. Trying to start pgpool anyway...\n")));
+				fprintf(stderr, "NOTICE: pid file found but it seems bogus. Trying to start pgpool anyway...\n");
 		}
 	}
 
@@ -337,10 +336,17 @@ main(int argc, char **argv)
 		char		dirnamebuf[POOLMAXPATHLEN + 1];
 		char	   *dirp;
 
-		strlcpy(dirnamebuf, conf_file, sizeof(dirnamebuf));
-		dirp = dirname(dirnamebuf);
-		snprintf(pool_passwd, sizeof(pool_passwd), "%s/%s",
-				 dirp, pool_config->pool_passwd);
+		if (pool_config->pool_passwd[0] != '/')
+		{
+			strlcpy(dirnamebuf, conf_file, sizeof(dirnamebuf));
+			dirp = dirname(dirnamebuf);
+			snprintf(pool_passwd, sizeof(pool_passwd), "%s/%s",
+					 dirp, pool_config->pool_passwd);
+		}
+		else
+			strlcpy(pool_passwd, pool_config->pool_passwd,
+				 sizeof(pool_passwd));
+
 		pool_init_pool_passwd(pool_passwd, POOL_PASSWD_R);
 	}
 
