@@ -4,8 +4,8 @@
  *	  implementation of Value nodes
  *
  *
- * Portions Copyright (c) 2003-2015, PgPool Global Development Group
- * Copyright (c) 2003-2015, PostgreSQL Global Development Group
+ * Portions Copyright (c) 2003-2019, PgPool Global Development Group
+ * Copyright (c) 2003-2019, PostgreSQL Global Development Group
  *
  *
  * IDENTIFICATION
@@ -23,7 +23,7 @@
  *	makeInteger
  */
 Value *
-makeInteger(long i)
+makeInteger(int i)
 {
 	Value	   *v = makeNode(Value);
 
