@@ -4,7 +4,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2018	PgPool Global Development Group
+ * Copyright (c) 2003-2019	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -207,6 +207,7 @@ pcp_worker_main(int port)
 
 		/* process a request */
 		pcp_process_command(tos, buf, rsize);
+		pfree(buf);
 	}
 	exit(0);
 }
@@ -871,6 +872,8 @@ inform_node_info(PCP_CONNECTION * frontend, char *buf)
 				  strlen(weight_str) + 1 +
 				  strlen(role_str) + 1 +
 				  strlen(standby_delay_str) + 1 +
+				  strlen(bi->replication_state) + 1 +
+				  strlen(bi->replication_sync_state) + 1 +
 				  strlen(status_changed_time_str) + 1 +
 				  sizeof(int));
 	pcp_write(frontend, &wsize, sizeof(int));
@@ -881,6 +884,8 @@ inform_node_info(PCP_CONNECTION * frontend, char *buf)
 	pcp_write(frontend, weight_str, strlen(weight_str) + 1);
 	pcp_write(frontend, role_str, strlen(role_str) + 1);
 	pcp_write(frontend, standby_delay_str, strlen(standby_delay_str) + 1);
+	pcp_write(frontend, bi->replication_state, strlen(bi->replication_state) + 1);
+	pcp_write(frontend, bi->replication_sync_state, strlen(bi->replication_sync_state) + 1);
 	pcp_write(frontend, status_changed_time_str, strlen(status_changed_time_str) + 1);
 
 	do_pcp_flush(frontend);
