@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2018	PgPool Global Development Group
+ * Copyright (c) 2003-2019	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -57,7 +57,7 @@
 #include "auth/md5.h"
 #include "auth/pool_passwd.h"
 #include "auth/pool_hba.h"
-
+#include "utils/pool_relcache.h"
 
 static StartupPacket *read_startup_packet(POOL_CONNECTION * cp);
 static POOL_CONNECTION_POOL * connect_backend(StartupPacket *sp, POOL_CONNECTION * frontend);
@@ -68,7 +68,7 @@ static RETSIGTYPE reload_config_handler(int sig);
 static RETSIGTYPE authentication_timeout(int sig);
 static void send_params(POOL_CONNECTION * frontend, POOL_CONNECTION_POOL * backend);
 static void send_frontend_exits(void);
-static void connection_count_up(void);
+static int	connection_count_up(void);
 static void connection_count_down(void);
 static bool connect_using_existing_connection(POOL_CONNECTION * frontend,
 								  POOL_CONNECTION_POOL * backend,
@@ -91,11 +91,15 @@ static void child_will_go_down(int code, Datum arg);
 static int opt_sort(const void *a, const void *b);
 
 /*
- * non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived
+ * Non 0 means SIGTERM (smart shutdown) or SIGINT (fast shutdown) has arrived
  */
 volatile sig_atomic_t exit_request = 0;
 static volatile sig_atomic_t alarm_enabled = false;
 
+/*
+ * Ignore SIGUSR1 if requested. Used when DROP DATABASE is requested.
+ */
+volatile sig_atomic_t ignore_sigusr1 = 0;
 
 static int	idle;				/* non 0 means this child is in idle state */
 static int	accepted = 0;
@@ -287,6 +291,7 @@ do_child(int *fds)
 		StartupPacket *sp;
 		int			front_end_fd;
 		SockAddr	saddr;
+		int			con_count;
 
 		/* reset per iteration memory context */
 		MemoryContextSwitchTo(ProcessLoopContext);
@@ -317,7 +322,32 @@ do_child(int *fds)
 		if (front_end_fd == RETRY)
 			continue;
 
-		connection_count_up();
+		/*
+		 * Check if max connections from clients execeeded.
+		 */
+		con_count = connection_count_up();
+		if (con_count > (pool_config->num_init_children - pool_config->reserved_connections))
+		{
+			POOL_CONNECTION * cp;
+			cp = pool_open(front_end_fd, false);
+			if (cp == NULL)
+			{
+				connection_count_down();
+				continue;
+			}
+			connection_count_down();
+			pool_send_fatal_message(cp, 3, "53300",
+									"Sorry, too many clients already",
+									"",
+									"",
+									__FILE__, __LINE__);
+			ereport(ERROR,
+					(errcode(ERRCODE_TOO_MANY_CONNECTIONS),
+					 errmsg("Sorry, too many clients already")));
+			pool_close(cp);
+			continue;
+		}
+
 		accepted = 1;
 
 		check_config_reload();
@@ -935,19 +965,20 @@ StartupPacketCopy(StartupPacket *sp)
 static POOL_CONNECTION_POOL * connect_backend(StartupPacket *sp, POOL_CONNECTION * frontend)
 {
 	POOL_CONNECTION_POOL *backend;
-	StartupPacket *topmem_sp;
+	StartupPacket *volatile topmem_sp = NULL;
+	volatile bool	topmem_sp_set = false;
 	int			i;
 
 	/* connect to the backend */
 	backend = pool_create_cp();
 	if (backend == NULL)
 	{
-		pool_send_error_message(frontend, sp->major, "XX000", "connection cache is full", "",
-								"increase max_pool", __FILE__, __LINE__);
+		pool_send_error_message(frontend, sp->major, "XX000", "all backend nodes are down, pgpool requires at least one valid node", "",
+								"repair the backend nodes and restart pgpool", __FILE__, __LINE__);
 		ereport(ERROR,
 				(errmsg("unable to connect to backend"),
-				 errdetail("connection cache is full"),
-				 errhint("increase the \"max_pool\" configuration value, current max_pool is %d", pool_config->max_pool)));
+				 errdetail("all backend nodes are down, pgpool requires at least one valid node"),
+				 errhint("repair the backend nodes and restart pgpool")));
 	}
 
 	PG_TRY();
@@ -975,6 +1006,7 @@ static POOL_CONNECTION_POOL * connect_backend(StartupPacket *sp, POOL_CONNECTION
 				 * save startup packet info
 				 */
 				CONNECTION_SLOT(backend, i)->sp = topmem_sp;
+				topmem_sp_set = true;
 
 				/* send startup packet */
 				send_startup_packet(CONNECTION_SLOT(backend, i));
@@ -998,10 +1030,17 @@ static POOL_CONNECTION_POOL * connect_backend(StartupPacket *sp, POOL_CONNECTION
 	PG_CATCH();
 	{
 		pool_discard_cp(sp->user, sp->database, sp->major);
+		topmem_sp = NULL;
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
 
+	/* At this point, we need to free previously allocated memory for the
+	 * startup packet if no backend is up.
+	 */
+	if (!topmem_sp_set && topmem_sp != NULL)
+		pfree(topmem_sp);
+
 	return backend;
 }
 
@@ -1076,6 +1115,12 @@ static RETSIGTYPE close_idle_connection(int sig)
 	ConnectionInfo *info;
 	int			save_errno = errno;
 
+	/*
+	 * DROP DATABSE is ongoing.
+	 */
+	if (ignore_sigusr1)
+		return;
+
 #ifdef NOT_USED
 	ereport(DEBUG1,
 			(errmsg("close connection request received")));
@@ -1250,7 +1295,8 @@ child_will_go_down(int code, Datum arg)
 	if (accepted)
 		connection_count_down();
 
-	if (pool_config->memory_cache_enabled && !pool_is_shmem_cache())
+	if ((pool_config->memory_cache_enabled || pool_config->enable_shared_relcache)
+		&& !pool_is_shmem_cache())
 	{
 		memcached_disconnect();
 	}
@@ -1490,10 +1536,11 @@ discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT * cp)
 }
 
 /*
- * Count up connection counter (from frontend to pgpool)
- * in shared memory
+ * Count up connection counter (from frontend to pgpool) in shared memory and
+ * returns current counter value.  Please note that the returned value may not
+ * be up to date since locking has been already released.
  */
-static void
+static int
 connection_count_up(void)
 {
 	pool_sigset_t oldmask;
@@ -1501,8 +1548,10 @@ connection_count_up(void)
 	POOL_SETMASK2(&BlockSig, &oldmask);
 	pool_semaphore_lock(CONN_COUNTER_SEM);
 	Req_info->conn_counter++;
+	elog(DEBUG5, "connection_count_up: number of connected children: %d", Req_info->conn_counter);
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
 	POOL_SETMASK(&oldmask);
+	return Req_info->conn_counter;
 }
 
 /*
@@ -1526,6 +1575,7 @@ connection_count_down(void)
 	 */
 	if (Req_info->conn_counter > 0)
 		Req_info->conn_counter--;
+	elog(DEBUG5, "connection_count_down: number of connected children: %d", Req_info->conn_counter);
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
 	POOL_SETMASK(&oldmask);
 }
@@ -1678,7 +1728,7 @@ select_load_balancing_node(void)
 
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
-		if (VALID_BACKEND(i))
+		if (VALID_BACKEND_RAW(i))
 		{
 			if (i == no_load_balance_node_id)
 				continue;
@@ -1704,7 +1754,7 @@ select_load_balancing_node(void)
 		if ((suggested_node_id == -1 && i == PRIMARY_NODE_ID) || i == no_load_balance_node_id)
 			continue;
 
-		if (VALID_BACKEND(i) && BACKEND_INFO(i).backend_weight > 0.0)
+		if (VALID_BACKEND_RAW(i) && BACKEND_INFO(i).backend_weight > 0.0)
 		{
 			if (r >= total_weight)
 				selected_slot = i;
@@ -1778,7 +1828,7 @@ check_restart_request(void)
 	if (pool_get_my_process_info()->need_to_restart)
 	{
 		ereport(LOG,
-				(errmsg("failback event detected"),
+				(errmsg("failover or failback event detected"),
 				 errdetail("restarting myself")));
 
 		pool_get_my_process_info()->need_to_restart = 0;
@@ -2081,7 +2131,7 @@ validate_backend_connectivity(int front_end_fd)
 		if (front_end_fd > 0)
 		{
 			POOL_CONNECTION *cp;
-			volatile StartupPacket *sp;
+			StartupPacket *volatile sp;
 
 			/*
 			 * we do not want to report socket error, as above errors will be
@@ -2105,15 +2155,17 @@ validate_backend_connectivity(int front_end_fd)
 										error_hint,
 										__FILE__,
 										__LINE__);
+				
 			}
 			PG_CATCH();
 			{
+				pool_free_startup_packet(sp);
 				FlushErrorState();
 				ereport(FATAL,
 						(errmsg("%s", error_msg), errdetail("%s", error_detail), errhint("%s", error_hint)));
-				pfree(sp);
 			}
 			PG_END_TRY();
+			pool_free_startup_packet(sp);
 		}
 		ereport(FATAL,
 				(errmsg("%s", error_msg), errdetail("%s", error_detail), errhint("%s", error_hint)));
@@ -2251,7 +2303,7 @@ retry_startup:
 	{
 		ereport(LOG,
 				(errmsg("selecting backend connection"),
-				 errdetail("failback event detected, discarding existing connections")));
+				 errdetail("failover or failback event detected, discarding existing connections")));
 
 		pool_get_my_process_info()->need_to_restart = 0;
 		close_idle_connection(0);
@@ -2422,3 +2474,156 @@ static int opt_sort(const void *a, const void *b)
 {
 	return strcmp( *(char **)a, *(char **)b);
 }
+
+/*
+ * Returns PostgreSQL version.
+ * The returned PgVersion struct is in static memory.
+ * Caller must not modify it.
+ *
+ * Note:
+ * Must be called while query context already exists.
+ * If there's something goes wrong, this raises FATAL. So never returns to caller.
+ *
+ */
+PGVersion *
+Pgversion(POOL_CONNECTION_POOL * backend)
+{
+#define VERSION_BUF_SIZE	10
+	static	PGVersion	pgversion;
+	static	POOL_RELCACHE *relcache;
+	char	*result;
+	char	*p;
+	char	buf[VERSION_BUF_SIZE];
+	int		i;
+	int		major;
+	int		minor;
+
+	/*
+	 * First, check local cache. If cache is set, just return it.
+	 */
+	if (pgversion.major != 0)
+	{
+		ereport(DEBUG5,
+				(errmsg("Pgversion: local cache returned")));
+
+		return &pgversion;
+	}
+
+	if (!relcache)
+	{
+		/*
+		 * Create relcache.
+		 */
+		relcache = pool_create_relcache(pool_config->relcache_size, "SELECT version()",
+										string_register_func, string_unregister_func, false);
+		if (relcache == NULL)
+		{
+			ereport(FATAL,
+					(errmsg("Pgversion: unable to create relcache while getting PostgreSQL version.")));
+			return NULL;
+		}
+	}
+
+	/*
+	 * Search relcache.
+	 */
+	result = (char *)pool_search_relcache(relcache, backend, "version");
+	if (result == 0)
+	{
+		ereport(FATAL,
+				(errmsg("Pgversion: unable to search relcache while getting PostgreSQL version.")));
+		return NULL;
+	}
+
+	ereport(DEBUG5,
+			(errmsg("Pgversion: version string: %s", result)));
+
+	/*
+	 * Extract major version number.  We create major version as "version" *
+	 * 10.  For example, for V10, the major version number will be 100, for
+	 * V9.6 it will be 96, and so on.  For alpha or beta version, the version
+	 * string could be something like "12beta1". In this case we assume that
+	 * atoi(3) is smart enough to stop at the first character which is not a
+	 * valid digit (in our case 'b')). So "12beta1" should be converted to 12.
+	 */
+	p = strchr(result, ' ');
+	if (p == NULL)
+	{
+		ereport(FATAL,
+				(errmsg("Pgversion: unable to find the first space in the version string: %s", result)));
+		return NULL;
+	}
+
+	p++;
+	i = 0;
+	while (i < VERSION_BUF_SIZE - 1 && p && *p != '.')
+	{
+		buf[i++] = *p++;
+	}
+	buf[i] = '\0';
+	major = atoi(buf);
+	ereport(DEBUG5,
+			(errmsg("Pgversion: major version: %d", major)));
+
+	/* Assuming PostgreSQL V100 is the final release:-) */
+	if (major < 6 || major > 100)
+	{
+		ereport(FATAL,
+				(errmsg("Pgversion: wrong major version: %d", major)));
+		return NULL;
+	}
+
+	/*
+	 * If major version is 10 or above, we are done to extract major.
+	 * Otherwise extract below decimal point part.
+	 */
+	if (major >= 10)
+	{
+		major *= 10;
+	}
+	else
+	{
+		p++;
+		i = 0;
+		while (i < VERSION_BUF_SIZE -1 && p && *p != '.' && *p != ' ')
+		{
+			buf[i++] = *p++;
+		}
+		buf[i] = '\0';
+		major = major * 10 + atoi(buf);
+		ereport(DEBUG5,
+				(errmsg("Pgversion: major version: %d", major)));
+		pgversion.major = major;
+	}
+
+	/*
+	 * Extract minor version.
+	 */
+	p++;
+	i = 0;
+	while (i < VERSION_BUF_SIZE -1 && p && *p != '.' && *p != ' ')
+	{
+		buf[i++] = *p++;
+	}
+	buf[i] = '\0';
+	minor = atoi(buf);
+	ereport(DEBUG5,
+			(errmsg("Pgversion: minor version: %d", minor)));
+
+	if (minor < 0 || minor > 100)
+	{
+		ereport(FATAL,
+				(errmsg("Pgversion: wrong minor version: %d", minor)));
+		return NULL;
+	}
+
+
+	/*
+	 * Ok, everything looks good. Set the local cache.
+	 */
+	pgversion.major = major;
+	pgversion.minor = minor;
+	strncpy(pgversion.version_string, result, sizeof(pgversion.version_string) - 1);
+
+	return &pgversion;
+}
