@@ -17,6 +17,8 @@ echo -n "creating test environment..."
 $PGPOOL_SETUP -m s -n 2 || exit 1
 echo "done."
 
+echo "search_primary_node_timeout = 5" >> etc/pgpool.conf
+
 source ./bashrc.ports
 
 export PGPORT=$PGPOOL_PORT
@@ -35,9 +37,17 @@ do
 	if [ $? = 0 ];then
 		break;
 	fi
+
+	grep "failover: no backends are degenerated" log/pgpool.log > /dev/null 2>&1
+	if [ $? = 0 ];then
+		break;
+	fi
+
 	sleep 1
 done
 
+sleep 5
+
 # if that's the case, shutdownall will hang.
 (./shutdownall)&
 sleep 5
