@@ -3,7 +3,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2018	PgPool Global Development Group
+ * Copyright (c) 2003-2019	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -227,6 +227,12 @@ pattern_compare(char *str, const int type, const char *param_name)
 	}
 
 	s = strip_quote(str);
+	if (!s)
+	{
+		elog(WARNING, "pattern_compare: strip_quote() returns error");
+		return -1;
+	}
+
 	for (i = 0; i < *pattc; i++)
 	{
 		if (lists_patterns[i].type != type)
@@ -271,13 +277,31 @@ pattern_compare(char *str, const int type, const char *param_name)
 	return result;
 }
 
+/*
+ * Returns double quotes stripped version of malloced string.
+ * Callers must free() after using it.
+ * Returns NULL on error.
+ */
 static char *
 strip_quote(char *str)
 {
 	char	   *after;
+	int		   len;
 	int			i = 0;
 
-	after = malloc(sizeof(char) * strlen(str) + 1);
+	len = strlen(str);
+	after = malloc(sizeof(char) * len + 1);
+	if (!after)
+	{
+		return NULL;
+	}
+
+	if (len == 0)
+	{
+		/* empty string case */
+		*after = '\0';
+		return after;
+	}
 
 	do
 	{
@@ -499,8 +523,7 @@ view_walker(Node *node, void *context)
 }
 
 /*
- * Judge the table used in a query represented by node is a system
- * catalog or not.
+ * Determine whether table_name is a system catalog or not.
  */
 static bool
 is_system_catalog(char *table_name)
@@ -519,9 +542,7 @@ is_system_catalog(char *table_name)
 
 #define ISBELONGTOPGCATALOGQUERY3 "SELECT count(*) FROM pg_class AS c, pg_namespace AS n WHERE c.oid = pg_catalog.to_regclass('\"%s\"') AND c.relnamespace = n.oid AND n.nspname = 'pg_catalog'"
 
-	int			hasreliscatalog;
 	bool		result;
-	static POOL_RELCACHE * hasreliscatalog_cache;
 	static POOL_RELCACHE * relcache;
 	POOL_CONNECTION_POOL *backend;
 
@@ -536,28 +557,9 @@ is_system_catalog(char *table_name)
 	backend = pool_get_session_context(false)->backend;
 
 	/*
-	 * Check if pg_namespace exists
+	 * Check if pg_namespace exists. PostgreSQL 7.3 or later has it.
 	 */
-	if (!hasreliscatalog_cache)
-	{
-		char	   *query;
-
-		query = HASPGNAMESPACEQUERY;
-
-		hasreliscatalog_cache = pool_create_relcache(pool_config->relcache_size, query,
-													 int_register_func, int_unregister_func,
-													 false);
-		if (hasreliscatalog_cache == NULL)
-		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking for system catalog")));
-			return false;
-		}
-	}
-
-	hasreliscatalog = pool_search_relcache(hasreliscatalog_cache, backend, "pg_namespace") == 0 ? 0 : 1;
-
-	if (hasreliscatalog)
+	if (Pgversion(backend)->major >= 73)
 	{
 		/*
 		 * If relcache does not exist, create it.
@@ -606,8 +608,7 @@ is_system_catalog(char *table_name)
 }
 
 /*
- * Judge the table used in a query represented by node is a temporary
- * table or not.
+ * Returns true if table_name is a temporary table.
  */
 static POOL_RELCACHE * is_temp_table_relcache;
 
@@ -635,41 +636,57 @@ is_temp_table(char *table_name)
  * name.
  */
 #define ISTEMPQUERY84 "SELECT count(*) FROM pg_catalog.pg_class AS c WHERE c.relname = '%s' AND c.relistemp"
-
-	int			hasrelistemp;
 	bool		result;
-	static POOL_RELCACHE * hasrelistemp_cache;
 	char	   *query;
 	POOL_CONNECTION_POOL *backend;
+	int			major;
 
-	if (table_name == NULL)
+	if (table_name == NULL || pool_config->check_temp_table == CHECK_TEMP_NONE ||
+		pool_config->check_temp_table == CHECK_TEMP_OFF)
 	{
 		return false;
 	}
 
 	backend = pool_get_session_context(false)->backend;
 
-	/*
-	 * Check backend version
-	 */
-	if (!hasrelistemp_cache)
+	if (pool_config->check_temp_table == CHECK_TEMP_TRACE)
 	{
-		hasrelistemp_cache = pool_create_relcache(pool_config->relcache_size, HASRELITEMPPQUERY,
-												  int_register_func, int_unregister_func,
-												  false);
-		if (hasrelistemp_cache == NULL)
+		POOL_TEMP_TABLE	*temp_table;
+
+		temp_table = pool_temp_tables_find(table_name);
+		if (temp_table && (temp_table->state == TEMP_TABLE_CREATE_COMMITTED ||
+						   temp_table->state == TEMP_TABLE_CREATING))
 		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking for temporary table")));
+			ereport(DEBUG1,
+					(errmsg("is_temp_table: %s is a temp table", table_name)));
+			return true;
+		}
+		else
+		{
+			ereport(DEBUG1,
+					(errmsg("is_temp_table: %s is not a temp table", table_name)));
 			return false;
 		}
 	}
 
-	hasrelistemp = pool_search_relcache(hasrelistemp_cache, backend, "pg_class") == 0 ? 0 : 1;
-	if (hasrelistemp)
+	/*
+	 * Below is check_temp_table == CHECK_TEMP_CATALOG or CHECK_TEMP_ON case.
+	 */
+
+	/*
+	 * Check backend version.
+	 */
+	major = Pgversion(backend)->major;
+	if (major == 73 || major == 83 || major >= 91)
+		query = ISTEMPQUERY83;
+	else if (major == 84 || major == 90)
 		query = ISTEMPQUERY84;
 	else
-		query = ISTEMPQUERY83;
+	{
+		ereport(WARNING,
+				(errmsg("is_temp_table: unexpected PostgreSQL version: %s", Pgversion(backend)->version_string)));
+		return false;	/* fall back to assume that the table is not a temporary table. */
+	}
 
 	/*
 	 * If relcache does not exist, create it.
@@ -708,8 +725,7 @@ discard_temp_table_relcache(void)
 }
 
 /*
- * Judge the table used in a query represented by node is a unlogged
- * table or not.
+ * Returns true if table_name is an unlogged table.
  */
 bool
 is_unlogged_table(char *table_name)
@@ -730,10 +746,9 @@ is_unlogged_table(char *table_name)
 
 #define ISUNLOGGEDQUERY3 "SELECT count(*) FROM pg_catalog.pg_class AS c WHERE c.oid = pg_catalog.to_regclass('%s') AND c.relpersistence = 'u'"
 
-	int			hasrelpersistence;
-	static POOL_RELCACHE * hasrelpersistence_cache;
 	static POOL_RELCACHE * relcache;
 	POOL_CONNECTION_POOL *backend;
+	int		major;
 
 	if (table_name == NULL)
 	{
@@ -746,23 +761,10 @@ is_unlogged_table(char *table_name)
 	backend = pool_get_session_context(false)->backend;
 
 	/*
-	 * Check backend version
+	 * Check backend version.
 	 */
-	if (!hasrelpersistence_cache)
-	{
-		hasrelpersistence_cache = pool_create_relcache(pool_config->relcache_size, HASRELPERSISTENCEQUERY,
-													   int_register_func, int_unregister_func,
-													   false);
-		if (hasrelpersistence_cache == NULL)
-		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking for unlogged table")));
-			return false;
-		}
-	}
-
-	hasrelpersistence = pool_search_relcache(hasrelpersistence_cache, backend, "pg_class") == 0 ? 0 : 1;
-	if (hasrelpersistence)
+	major = Pgversion(backend)->major;
+	if (major >= 91)
 	{
 		bool		result;
 		char	   *query;
@@ -811,7 +813,8 @@ is_unlogged_table(char *table_name)
 }
 
 /*
- * Judge the table used in a query is a view or not.
+ * Returns true if table_name is a view.
+ * This function is called by query cache module.
  */
 bool
 is_view(char *table_name)
@@ -913,39 +916,22 @@ pool_has_pgpool_regclass(void)
 }
 
 /*
- * Judge if we have to_regclass or not.
+ * Returns true if we have to_regclass().
  */
 bool
 pool_has_to_regclass(void)
 {
-/*
- * Query to know if to_regclass exists.
- */
-#define HAS_TOREGCLASSQUERY "SELECT count(*) from (SELECT has_function_privilege('%s', 'pg_catalog.to_regclass(cstring)', 'execute') WHERE EXISTS(SELECT * FROM pg_catalog.pg_proc AS p WHERE p.proname = 'to_regclass')) AS s"
-
-	bool		result;
-	static POOL_RELCACHE * relcache;
 	POOL_CONNECTION_POOL *backend;
-	char	   *user;
+	PGVersion	*pgversion;
 
 	backend = pool_get_session_context(false)->backend;
-	user = MASTER_CONNECTION(backend)->sp->user;
+	pgversion = Pgversion(backend);
 
-	if (!relcache)
-	{
-		relcache = pool_create_relcache(pool_config->relcache_size, HAS_TOREGCLASSQUERY,
-										int_register_func, int_unregister_func,
-										false);
-		if (relcache == NULL)
-		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking to_regclass presence")));
-			return false;
-		}
-	}
+	/* PostgreSQL 9.4 or above has to_regclass() */
+	if (pgversion->major >= 94)
+		return true;
 
-	result = pool_search_relcache(relcache, backend, user) == 0 ? 0 : 1;
-	return result;
+	return false;
 }
 
 /*
